/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/module_table.h"
#include <stdio.h>
#include <cstring>

// ##################################################################
/**
 * @brief Initialisieren der Membervariablen.
 *
 * @param id Nummer der Kanalklasse fuer z.B. Trace.
 */
Datafox::Communication::Module_table::Module_table( int connection_id, uint8_t device_address )
    : Module_application( connection_id, device_address ) 
{
    trace( "%s(%i):INFO(3): %s [ %i, %u ]\n", __FILE__, __LINE__, __func__, connection_id, device_address );
}

// ##################################################################
/**
 * @brief Destruktor.
 *
 */
Datafox::Communication::Module_table::~Module_table() 
{
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void Datafox::Communication::Module_table::error_module( uint16_t module, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s [ %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, module, error, text.c_str() );

    // Ist IV_MODULE_TABLE ?
    if ( module == IV_MODULE_TABLE )
    {

    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    Module_application::error_module( module, error, text );
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void Datafox::Communication::Module_table::error_function( uint16_t module, uint16_t func, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s [ %u, %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, module, func, error, text.c_str() );

    if ( module == IV_MODULE_TABLE )
    {
        // Sonderfehler für Passwortschutz abhandeln.
        if ( error == 28 )
        {
            // Die Modulfunktionen liefern den Fehlercode 28, wenn kein oder ein falsches Passwort verwendet wurde.
            // Es ist in diesem Fall zuerst das richtige Passwort für die Kommunikation erforderlich.
            text = boost::str( boost::format( "( %u ): Korrektes Passwort für die Kommunikation erforderlich." ) % error );
            return;
        }
        else if ( error == 29 )
        {
            // Die Modulfunktionen liefern den Fehlercode 29, wenn 3 mal ein falsches Passwort verwendet wurde.
            // Die Befehlsausführung ist danach für 60 Sekunden blockiert.
            text = boost::str( boost::format( "( %u ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt." ) % error );
            return;
        }

        switch ( func )
        {
        case IV_FUNCTION_TABLE_GET:
        case IV_FUNCTION_TABLE_GET_COLUMN:
            switch ( error )
            {
            case 1: text = boost::str( boost::format( "( %u ): Der Majorbestandteil des Tabellennamens ist unbekannt." ) % error ); break;
            case 2: text = boost::str( boost::format( "( %u ): Der Minorbestandteil des Tabellennamens ist unbekannt." ) % error ); break;
            case 3: text = boost::str( boost::format( "( %u ): Der Subminorbestandteil (Feldname) des Tabellenfeldnamens ist unbekannt." ) % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_OPEN:
            switch ( error )
            {
            case 1: text = boost::str( boost::format( "( %u ): Der Majorbestandteil des Tabellennamens ist unbekannt." ) % error ); break;
            case 2: text = boost::str( boost::format( "( %u ): Der Minorbestandteil des Tabellennamens ist unbekannt." ) % error ); break;
                // Standardmaessig ist die Option mit 0 angegeben. Ein gesetztes Bit, dessen Bedeutung unklar ist, ergibt diesen Fehler.
            case 3: text = boost::str( boost::format( "( %u ): Unbekannte Option gesetzt." ) % error ); break;
                // Wenn ein reservierter Parameter angegeben wurde.
            case 4: text = boost::str( boost::format( "( %u ): Der angegebene Reservierte Parameter wird nicht unterstützt." ) % error ); break;
                // Die Tabelle ist bereits mit schreib - Recht geoffnet worden.
            //case 5: text = boost::str( boost::format("( %u ): Die Tabelle kann nur lesend geoeffnet werden." ) % error ); break;
                // Die Tabelle ist bereits mit Exclusive - Recht geoffnet worden.
            //case 6: text = boost::str( boost::format("( %u ): Die Tabelle ist z. Z. gesperrt." ) % error ); break;
                // Die Nutzung der Tabelle wurde aufgrund von Resourcenmangel versagt. (Zu oft geoeffnet worden usw..)
            //case 7: text = boost::str( boost::format("( %u ): Die Nutzung der Tabelle wurde aufgrund von Resourcenmangel versagt." ) % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_CLOSE:
            switch ( error )
            {
                // Wenn versucht wird ein Handle zu schliessen der keiner offenen Tabelle zugehoerig ist.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle war nicht geoffnet." ) % error ); break;
                // Liste konnte nicht geschlossen werden.
            case 2: text = boost::str( boost::format( "( %u ): Tabelle konnte nicht geschlossen werden." ) % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_SET_FILTER:
        case IV_FUNCTION_TABLE_SET_FILTER_OK:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Ungueltiger Selektionsspaltenanzahl.
            case 2: text = boost::str( boost::format( "( %u ): Falsche Anzahl von Selektionsspalten." ) % error ); break;
                // Nach setzen des Filterkriteriums -> Kein Datensatz
            case 3: text = boost::str( boost::format( "( %u ): Filterkriterium liefert kein Ergebnis." ) % error ); break;
                // Spaltenname des Filter unbekannt.
            case 4: text = boost::str( boost::format( "( %u ): Unbekannter Spaltenname." ) % error ); break;
                // Unbekannter (nicht unterstuetzter) Filteroperator.
            case 5: text = boost::str( boost::format( "( %u ): Unbekannter Filteroperator." ) % error ); break;
                // Fehlender in Anfuehrungszeichenpaar des zu filterndem Spaltenwerts.
                // Keine vorhanden, unterschiedliches ein/aus '" oder fehlender ein oder Ausgang.
            case 6: text = boost::str( boost::format( "( %u ): Fehler in Anfuehrungszeichenpaar des zu filterndem Spaltenwertes." ) % error ); break;
                // Zu viele Parameter angegeben.
            case 7: text = boost::str( boost::format("( %u ): Zu viele Parameter angegeben." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_GET_FILTER:
        case IV_FUNCTION_TABLE_REMOVE_FILTER:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Auf der angegebenen Tabelle ist kein Filter gesetzt.
            //case 2:text = boost::str( boost::format("( %u ): Kein Filter bei Tabelle angegeben." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_SET_CURRENT_ROW:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Wert fuer Origin ausserhalb des gueltigen Bereichs. (0, 1, 2)
            case 2: text = boost::str( boost::format( "( %u ): Wert fuer Origin ungueltig." ) % error ); break;
                // Sprungweite (Offset) fuehrte zu BOF (begin of file). Der Datensatzzeiger steht damit vor dem ersten Datensatz.
            case 3: text = boost::str( boost::format( "( %u ): Sprung fuehrte zu BOF." ) % error ); break;
                // Sprungweite (Offset) fuehrte zu EOF (end of file). Der Datensatzzeiger steht damit hinter dem letzten Datensatz.
            case 4: text = boost::str( boost::format( "( %u ): Sprung fuehrte zu EOF." ) % error ); break;
                // Sprung konnte aufgrund von angegebenen offset mit origin nicht durchgefuehrt werden.
                // Z. B. Filter gesetzt und Sprung auf aktuellen der nicht dem Filterkriterium entspraeche.
            case 5: text = boost::str( boost::format( "( %u ): Sprung konnte nicht durchgefuehrt werden." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA_1:
        case IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Spaltenname wurde nicht gefunden.
            case 2: text = boost::str( boost::format( "( %u ): Spaltenname unbekannt." ) % error ); break;
                // Angegebener Wert ist zu lang
            case 3: text = boost::str( boost::format( "( %u ): Zu setzender Wert ist zu lang." ) % error ); break;
                // Formatfehler in zu setzendem Spaltenwert.
            case 4: text = boost::str( boost::format( "( %u ): Formatfehler in zu setzendem Spaltenwert." ) % error ); break;
                // Schreiben des Spaltenwerts fehlgeschlagen
            case 5: text = boost::str( boost::format( "( %u ): Schreiben des Spaltenwerts fehlgeschalgen." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

         case IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA_1:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Fehler beim Oeffnen der Datei.
            case 2: text = boost::str( boost::format( "( %u ): Fehler beim Oeffnen der Datei." ) % error ); break;
                // Datei ist Leer.
            case 3: text = boost::str( boost::format( "( %u ): Keine Daten verfuegbar." ) % error ); break;
                // Sprungweite (Offset) fuehrte zu EOF (end of file). Der Datensatzzeiger steht damit hinter dem letzten Datensatz.
            case 4: text = boost::str( boost::format( "( %u ): Sprung fuehrte zu EOF." ) % error ); break;
                // Fehler beim Lesen der Datenzeile
            case 24: text = boost::str( boost::format( "( %u ): Fehler beim Lesen der Datenzeile." ) % error ); break;
                // Datenzeile ist als geloescht markiert
            case 26: text = boost::str( boost::format( "( %u ): Datenzeile ist als geloescht markiert." ) % error ); break;
                // Allgemeiner Fehler
            case 130:
            case 131: text = boost::str( boost::format( "( %u ): Allgemeiner Fehler." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;
            break;

        case IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA_1:
        case IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA:
        case IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA_1:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Fehler beim Oeffnen der Datei.
            case 2: text = boost::str( boost::format( "( %u ): Fehler beim Oeffnen der Datei." ) % error ); break;
                // Datei ist Leer.
            case 3: text = boost::str( boost::format( "( %u ): Keine Daten verfuegbar." ) % error ); break;
                // Sprungweite (Offset) fuehrte zu EOF (end of file). Der Datensatzzeiger steht damit hinter dem letzten Datensatz.
            case 4: text = boost::str( boost::format( "( %u ): Sprung fuehrte zu EOF." ) % error ); break;
                // Fehler beim Lesen der Datenzeile
            case 24: text = boost::str( boost::format( "( %u ): Fehler beim Lesen der Datenzeile." ) % error ); break;
                // Datenzeile ist als geloescht markiert
            case 26: text = boost::str( boost::format( "( %u ): Datenzeile ist als geloescht markiert." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Fehler kein Speicher mehr.
            case 4: text = boost::str( boost::format( "( %u ): Fehler beim reservieren von Speicher." ) % error ); break;
                // Fehler beim Lesen der Datenzeile.
            case 20: text = boost::str( boost::format( "( %u ): Feghler beim Lesend er Datenzeile." ) % error ); break;
                // Sprungweite (Offset) fuehrte zu EOF (end of file). Der Datensatzzeiger steht damit hinter dem letzten Datensatz.
            case 21: text = boost::str( boost::format( "( %u ): Sprung fuehrte zu EOF." ) % error ); break;
                // Fehler beim Lesen der Datenzeile
            case 22: text = boost::str( boost::format( "( %u ): Fehler beim Lesen der Datenzeile." ) % error ); break;
                // Allgemeiner Fehler beim Lesen der Datenzeile
            case 23: text = boost::str( boost::format( "( %u ): Allgemeiner Fehler beim Lesen der Datenzeile." ) % error ); break;
                // Datenzeile ist als geloescht markiert
            case 26: text = boost::str( boost::format( "( %u ): Datenzeile ist als geloescht markiert." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_GENERATE:
        case IV_FUCNTION_TABLE_GENERATE_OK:
            switch( error )
            {
            case 7: text = boost::str( boost::format( "( %u ): Das Modul kann zur Zeit nicht angesprochen werden." ) % error ); break;
            case 8: text = boost::str( boost::format( "( %u ): Fuer die erforderliche Funktion steht nicht ausreichend Speicher zur verfuegung." ) % error ); break;
            case 129: text = boost::str( boost::format( "( %u ): Das Modul bringt einen nicht näher spezifizierten Fehler." ) % error ); break;
            case 133: text = boost::str( boost::format( "( %u ): Das Modul enthält keine Daten zum erzeugen einer Liste." ) % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;
            

        case IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD:
        case IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD_OK:
        case IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD1:
        case IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD:
        case IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD_OK:
        case IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD1:
        case IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD:
        case IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD_OK:
            switch ( error )
            {
            case 1: text = boost::str( boost::format( "( %u ): Das gewählte Templateformat ist unbekannt." ) % error ); break;
            case 2: text = boost::str( boost::format( "( %u ): Das Modul antwortet nicht auf die Anfrage." ) % error ); break;
            case 3: text = boost::str( boost::format( "( %u ): Das gewählte Templateformat entspricht nicht dem verwendeten." ) % error ); break;
            case 7: text = boost::str( boost::format( "( %u ): Das Modul kann zur Zeit nicht angesprochen werden." ) % error ); break;
            case 129: text = boost::str( boost::format( "( %u ): Das Modul bringt einen nicht näher spezifizierten Fehler." ) % error ); break;
            case 132: text = boost::str( boost::format( "( %u ): Die Datenbank des Moduls ist voll." ) % error ); break;
            case 133: text = boost::str( boost::format( "( %u ): Das Modul enthält keine Fingertemplates." ) % error ); break;
            case 134: text = boost::str( boost::format( "( %u ): Das Fingertemplate existiert bereits." ) % error ); break;
            case 135: text = boost::str( boost::format( "( %u ): Das angeforderte Fingertemplate existiert nicht." ) % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_GET_ROW_COUNT:
        case IV_FUNCTION_TABLE_GET_CURRENT_ROW:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Fehler im Funktionsaufruf.
            case 255: text = boost::str( boost::format( "( %u ): Fehler im Funktionsaufruf." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_APPEND_ROW_DATA:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Fehler kein Speicher verfuegbar.
            case 4: text = boost::str( boost::format( "( %u ): Fehler kein Speicher verfuegbar." ) % error ); break;
                // Fehler bei Funktionsausfuehrung.
            case 3:
            case 5:
            case 7:
            case 9: 
            case 10: text = boost::str( boost::format( "( %u ): Fehler bei Funktionsausfuehrung." ) % error ); break;
                // Datenzeile ist als geloescht markiert
            case 26: text = boost::str( boost::format( "( %u ): Datenzeile ist als geloescht markiert." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_DELETE_CURRENT_ROW:
        case IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS:
        case IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS_OK:
           switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Sprungweite (Offset) fuehrte zu EOF (end of file). Der Datensatzzeiger steht damit hinter dem letzten Datensatz.
            case 14: text = boost::str( boost::format( "( %u ): Sprung fuehrte zu EOF." ) % error ); break;
                // Allgemeiner Fehler beim loeschen der Datenzeile
            case 12:
            case 13:
            case 30: text = boost::str( boost::format( "( %u ): Allgemeiner Fehler beim loeschen der Datenzeile." ) % error ); break;
                // Kein Filter aktiv.
            case 31: text = boost::str( boost::format( "( %u ): Kein Filter aktiv." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA:
           switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Fehler kein Speicher verfuegbar.
            case 4: text = boost::str( boost::format( "( %u ): Fehler kein Speicher verfuegbar." ) % error ); break;
                // Handlefehler
            case 15:
            case 16: text = boost::str( boost::format( "( %u ): Handlefehler." ) % error ); break;
            case 17: text = boost::str( boost::format( "( %u ): Dateifehler." ) % error ); break;
            case 18:
            case 19: text = boost::str( boost::format( "( %u ): Lesefehler." ) % error ); break;
                // Datenzeile ist als geloescht markiert
            case 26: text = boost::str( boost::format( "( %u ): Datenzeile ist als geloescht markiert." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        case IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA:
        case IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA_OK:
            switch ( error )
            {
                // Wenn der angegebene Handle nicht existiert.
            case 1: text = boost::str( boost::format( "( %u ): Angegebener Handle ist nicht geoffnet." ) % error ); break;
                // Spaltenname wurde nicht gefunden.
            case 2: text = boost::str( boost::format( "( %u ): Spaltenname unbekannt." ) % error ); break;
                // Angegebener Wert ist zu lang
            case 3: text = boost::str( boost::format( "( %u ): Zu setzender Wert ist zu lang." ) % error ); break;
                // Formatfehler in zu setzendem Spaltenwert.
            case 4: text = boost::str( boost::format( "( %u ): Formatfehler in zu setzendem Spaltenwert." ) % error ); break;
                // Schreiben des Spaltenwerts fehlgeschlagen
            case 5: text = boost::str( boost::format( "( %u ): Schreiben des Spaltenwerts fehlgeschalgen." ) % error ); break;

            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!" ) % error ); break;
            }
            break;

        default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Fehlerblock für Funktion!" ) % error ); break;
        }

        return;
    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    Module_application::error_function( module, func, error, text );
}

//#########################################################
//####################### TABLE ###########################
//#########################################################

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::remote_call( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count )
{
    if ( module == IV_MODULE_TABLE )
    {
        switch ( func )
        {
        // Ermitteln der Tabellendefinition
        case IV_FUNCTION_TABLE_GET: 
            return read_table_struct( get_device_address_sender(), reinterpret_cast<char *>( data ), reinterpret_cast<StructTable2 &>( data[strlen(reinterpret_cast<char *>( data )) + 1] ) );

        // Ermitteln der Felddefinition
        case IV_FUNCTION_TABLE_GET_COLUMN: 
            return read_table_column_struct( get_device_address_sender(), reinterpret_cast<char *>( data ), reinterpret_cast<StructTable2Field &>( data[strlen( reinterpret_cast<char *>( data )) + 1] ) ); 

        case IV_FUNCTION_TABLE_OPEN:
        case IV_FUNCTION_TABLE_CLOSE:
        case IV_FUNCTION_TABLE_SET_FILTER:
        case IV_FUNCTION_TABLE_SET_FILTER_OK:
        case IV_FUNCTION_TABLE_GET_FILTER:
        case IV_FUNCTION_TABLE_REMOVE_FILTER:
        case IV_FUNCTION_TABLE_SET_CURRENT_ROW:
        case IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA:
        case IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA:
        case IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA:
        case IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA:
        case IV_FUNCTION_TABLE_GENERATE:
            // Funktion nicht definiert
            return 0x0200; 

        case IV_FUCNTION_TABLE_GENERATE_OK:
            return reply_generate_table( get_device_address_sender() );

        case IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD:
            // Funktion nicht definiert
            return 0x0200; 

        case IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD_OK:
            return reply_append_fingerprint_record( get_device_address_sender() );

        case IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD:
            // Funktion nicht definiert
            return 0x0200; 

        case IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD_OK:
            return reply_read_fingerprint_record( get_device_address_sender() );

        case IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD:
            // Funktion nicht definiert
            return 0x0200; 

        case IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD_OK:
            return reply_delete_fingerprint_record( get_device_address_sender() );

        case IV_FUNCTION_TABLE_GET_ROW_COUNT:
        case IV_FUNCTION_TABLE_GET_CURRENT_ROW:
            // Funktion nicht definiert
            return 0x0200;

        case IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA:
            return reply_write_all_table_rows_column_data(get_device_address_sender());

        case IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS:
            return reply_delete_filtrated_table_rows(get_device_address_sender());

        // Funktion nicht definiert
        default: return 0x0200; 
        }

    }

    return Module_application::remote_call( module, func, data, count );
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_table_struct( unsigned char device_address, const std::string &name, StructTable2 &table )
{
    trace( "%s(%i):INFO(3): %s [ %i, %s, STable2 ]\n", __FILE__, __LINE__, __func__, device_address, name.c_str() );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Name als Parameter eintragen
        std::copy(name.cbegin(), name.cend(), &m_packet_data[0]);
        count += static_cast<int>( name.size() );
        m_packet_data[count] = 0; // Nullterminierung
        ++count;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET, result );
            return result;
        }

        // OK
        if ( count != sizeof( StructTable2 ) )
        {
            trace( "%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Struktur uebernehmen
        table = *reinterpret_cast<StructTable2 *>( m_packet_data );
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char data[256];
        StructTable2 t;
        count = 0;
        short err = IV_ERROR_SUCCESS;

        strcpy( reinterpret_cast<char *>( data ), name.c_str() );
        memset( &t, 0, sizeof( StructTable2 ) );

        char *major;
        char *minor;

        major = strtok( reinterpret_cast<char *>( data ), "." );
        minor = strtok( NULL, "." );

        if ( strcmp( major, "RECORD" ) == 0 )
        {
            switch ( atoi( minor ) )
            {
            case 0:
                t.num = 0;
                strcpy( t.name, "PZE-Meldung" );
                t.count = 5;
                break;

            case 1:
                t.num = 1;
                strcpy( t.name, "Status" );
                t.count = 8;
                break;

            default:
                // Minorbestandteil undefiniert
                err = 0x0002;
            }
        }
        else if ( strstr( reinterpret_cast<char *>( data ), "LIST" ) == reinterpret_cast<char *>( data ) )
        {
                // Minorbestandteil undefiniert
                err = 0x0002;
        }
        else if (strstr(reinterpret_cast<char *>( data ), "ACCESS") == reinterpret_cast<char *>( data ) )
        {
                // Minorbestandteil undefiniert
                err = 0x0002;
        }
        else
        {
            // Majorbestandteil undefiniert.
            err = 0x0001;
        }

        if (err == IV_ERROR_SUCCESS)
        {
            memcpy( &data[count], &t, sizeof( StructTable2 ) );
            count = sizeof( StructTable2 );
        }
        else
        {
            *reinterpret_cast<short *>( data ) = err;
            count = sizeof( short );
        }
        // Antwort des Befehls senden.
        if ( (result = call( (err == IV_ERROR_SUCCESS) ? IV_CONTEXT_TRANSMIT_ANSWER : IV_CONTEXT_TRANSMIT_ANSWER_ERROR, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET, data, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_table_column_struct( unsigned char device_address, const std::string &name, StructTable2Field &field )
{
    trace( "%s(%i):INFO(3): %s [ %i, %s, StructTable2Field ]\n", __FILE__, __LINE__, __func__, device_address, name.c_str() );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Name als Parameter eintragen
        std::copy(name.cbegin(), name.cend(), &m_packet_data[0]);
        count += static_cast<int>( name.size() );
        m_packet_data[count] = 0; // Nullterminierung
        ++count;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_COLUMN, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_COLUMN, result );
            return result;
        }

        // OK
        if ( count != sizeof( StructTable2Field ) )
        {
            trace( "%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Sturktur uebernehmen
        field = *reinterpret_cast<StructTable2Field *>( m_packet_data );
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char data[256];
        StructTable2Field f;
        count = 0;
        short err = IV_ERROR_SUCCESS;

        strcpy( reinterpret_cast<char *>( data ), name.c_str() );
        memset( &field, 0, sizeof( field ) );
        char *major;
        char *minor;
        char *subminor;

        major = strtok( reinterpret_cast<char *>( data ), "." );
        minor = strtok( NULL, "." );
        subminor = strtok( NULL, "." );

        if ( strcmp( major, "RECORD" ) == 0 )
        {
            switch ( atoi( minor ) )
            {
            case 0:
                {
                switch ( atoi( subminor ) )
                {
                case 0:
                    f.num = 0;
                    strcpy( f.name, "DU" );
                    f.type = TF_DT;
                    f.size = 7;
                    break;

                case 1:
                    f.num = 1;
                    strcpy( f.name, "Kennung" );
                    f.type = TF_A_NUM;
                    f.size = 5;
                    break;

                case 2:
                    f.num = 2;
                    strcpy( f.name, "PID" );
                    f.type = TF_A_NUM;
                    f.size = 9;
                    break;

                case 3:
                    f.num = 3;
                    strcpy( f.name, "IP" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;

                case 4:
                    f.num = 4;
                    strcpy( f.name, "Port" );
                    f.type = TF_A_NUM;
                    f.size = 9;
                    break;

                default:
                    err = 0x0003;
                }
                break;
                }
            case 1:
                {
                switch ( atoi( subminor ) )
                {
                case 0:
                    f.num = 0;
                    strcpy( f.name, "DU" );
                    f.type = TF_DT;
                    f.size = 7;
                    break;

                case 1:
                    f.num = 1;
                    strcpy( f.name, "Feld1" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;

                case 2:
                    f.num = 2;
                    strcpy( f.name, "Feld2" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;

                case 3:
                    f.num = 3;
                    strcpy( f.name, "Feld3" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;

                case 4:
                    f.num = 4;
                    strcpy( f.name, "Feld4" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;

                case 5:
                    f.num = 5;
                    strcpy( f.name, "Feld5" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;
                case 6:
                    f.num = 6;
                    strcpy( f.name, "Feld6" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;

                case 7:
                    f.num = 7;
                    strcpy( f.name, "Feld7" );
                    f.type = TF_A_NUM;
                    f.size = 21;
                    break;

                default:
                    err = 0x0003;
                }
                break;}
            default:
                err = 0x0002;
            }
        }
        else if ( strcmp( major, "LIST" ) == 0 )
        {
                err = 0x0002;
        }
        else if ( strcmp( reinterpret_cast<char *>( data ), "ACCESS") == 0 )
        {
                err = 0x0002;
        }
        else
        {
            // Majorbestandteil undefiniert.
            err = 0x0001;
        }

        if ( err == IV_ERROR_SUCCESS )
        {
            memcpy( &data[count], &f, sizeof( StructTable2Field ) );
            count = sizeof( StructTable2Field );
        }
        else
        {
            *reinterpret_cast<short *>( data ) = err;
            count = sizeof( short );
        }

        // Antwort des Befehls senden.
        if ( (result = call( (err == IV_ERROR_SUCCESS) ? IV_CONTEXT_TRANSMIT_ANSWER : IV_CONTEXT_TRANSMIT_ANSWER_ERROR, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_COLUMN, data, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

/*! 
    \brief Oeffnen einer Tabelle zum Bearbeiten.
    
    \param[in] device_address Anzusprechende Geraetenummer.
    \param[in] name Voll qualifizierter Name der Tabelle.
    \param[in] options Zusatzoptionen. Derzeit nicht verwendet.
    \param[out] handle Handle auf die geoeffnete Tabelle.
    \param[in, out] reserved Zeichenfolge die zur Uebertragung einer aktuell nicht verwendeten Zeichenfolge reserviert ist.
    \param[in, out] reservedSize Als Eingabe die Länge des reservierten Speichers für reserved, als Ausgabe die an reserved uebergebene Zeichenfolge.

    \retval 0 Ausfuehrung erfolgreich.
*/
int Datafox::Communication::Module_table::open_table(unsigned char device_address, const std::string &name, unsigned int options, int &handle, std::string &reserved)
{
    trace( "%s(%i):INFO(3): %s [ %i, %s, %i, %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, device_address, name.c_str(), options, handle, reserved.c_str() );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Parameter eintragen
        std::copy(name.cbegin(), name.cend(), &m_packet_data[count]);
        count += static_cast<int>( name.size() );
        m_packet_data[count] = 0; // Nullterminierung
        count++;

        *reinterpret_cast<unsigned int *>(&m_packet_data[count]) = options;
        count += sizeof(unsigned int);

        std::copy(reserved.cbegin(), reserved.cend(), &m_packet_data[count]);
        count += static_cast<int>( reserved.size() );
        m_packet_data[count] = 0; // Nullterminierung
        count++;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_OPEN, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_OPEN, result );
            return result;
        }

        // OK
        // Ein leerbyte fuer reservierten String in Antwort
        if (count != (sizeof(int) + sizeof(unsigned char)))
        {
            trace( "%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Handle uebernehmen
        handle = *reinterpret_cast<int*>(m_packet_data);
        // Eventuell zurueckgegebene Zeichenfolge uebernehmen.
        reserved.resize(count - sizeof(int));
        std::copy(&m_packet_data[sizeof(int)], &m_packet_data[count], reserved.begin());
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_OPEN, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif

    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::close_table( unsigned char device_address, int handle )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );


        // Parameter eintragen
        *reinterpret_cast<int*>( m_packet_data ) = handle;
        count = sizeof( int );

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_CLOSE, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_CLOSE, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_CLOSE, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif

    }

    return result;
}


// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::write_table_filter( unsigned char device_address, int handle, const std::string &filter )
{
    trace( "%s(%i):INFO(3): %s [ %i, 0x%08X ]\n", __FILE__, __LINE__, __func__, device_address, filter.c_str() );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );



        // Parameter eintragen
        *reinterpret_cast<int*>( m_packet_data ) = handle;
        count += sizeof( int );
        // Filter uebernehmen
        std::copy(filter.cbegin(), filter.cend(), &m_packet_data[count]);
        count += static_cast<int>( filter.size() );
        m_packet_data[count] = 0; // Nullterminierung
        ++count;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_FILTER, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_FILTER, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace("%s(%i):WARNING(3): %s Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count);
        }

        if ( (result = call( IV_CONTEXT_SLAVE, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_FILTER_OK, m_packet_data, count, IV_PENDING_TRIALS( get_connection().get_read_timeout() ) )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_FILTER_OK, result );
            return result;
        }

        // Keine Aktive - Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s.Pending [ Aktive - Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_FILTER, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif

    }


    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::reply_write_table_filter( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace( "%s(%i):ERROR(1): %s", __FILE__, __LINE__, __func__ );
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_FILTER_OK, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }


    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_table_filter( unsigned char device_address, int handle, std::string &filter )
{
    trace( "%s(%i):INFO(3): %s [ %i, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, filter.c_str() );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Parameter uebernehmen
        *reinterpret_cast<int*>( m_packet_data ) = handle;
        count = sizeof(int);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_FILTER, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_FILTER, result );
            return result;
        }

        // OK
        if ( count > 80 )
        {
            trace( "%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Wert uebernehmen
        filter = reinterpret_cast<char *>(&m_packet_data[0]);
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_FILTER, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif

    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::remove_table_filter( unsigned char device_address, int handle )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );


        // Name als Parameter eintragen
        *reinterpret_cast<int*>( m_packet_data ) = handle;
        count = sizeof(int);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_REMOVE_FILTER, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_REMOVE_FILTER, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_REMOVE_FILTER, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif

    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_table_row_count( unsigned char device_address, int handle, unsigned int &current_count, unsigned int &unsorted_count, unsigned int &deleted_count )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, %i, %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, current_count, unsorted_count, deleted_count );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Handle als Parameter eintragen
        *reinterpret_cast<int *>(&m_packet_data[count]) = handle;
        count = sizeof(int);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_ROW_COUNT, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_ROW_COUNT, result );
            return result;
        }

        // OK
        if (count > static_cast<int>( sizeof(unsigned int) + sizeof(unsigned int) + sizeof(unsigned int) ))
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Anzahl uebernehmen
        current_count = *reinterpret_cast<unsigned int *>(m_packet_data);
        // Ab Firmware 04.03.08 werden zusaetzliche Parameter geliefert.
        if (count > static_cast<int>( sizeof(unsigned int) ))
        {
            unsorted_count = *reinterpret_cast<unsigned int *>(&m_packet_data[sizeof(unsigned int)]);
            deleted_count = *reinterpret_cast<unsigned int *>(&m_packet_data[sizeof(unsigned int) * 2]);
        }
        else
        {
            unsorted_count = 0;
            deleted_count = 0;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_ROW_COUNT, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_current_table_row( unsigned char device_address, int handle, unsigned int &current_row )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, current_row );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );


        // Name als Parameter eintragen
        *reinterpret_cast<int*>( &m_packet_data[count] ) = handle;
        count = sizeof(int);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW, result );
            return result;
        }

        // OK
        if ( count > static_cast<int>( sizeof(unsigned int) ) )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Anzahl uebernehmen
        current_row = *reinterpret_cast<unsigned int *>(m_packet_data);
    }
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::write_current_table_row( unsigned char device_address, int handle, int offset, unsigned short direction )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, offset, direction );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );


        // Parameter eintragen
        *reinterpret_cast<int *>(m_packet_data) = handle;
        count = sizeof(int);
        *reinterpret_cast<int *>(&m_packet_data[count]) = offset;
        count += sizeof(int);
        *reinterpret_cast<unsigned short *>(&m_packet_data[count]) = direction;
        count += sizeof(unsigned short);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::write_current_table_row_data( unsigned char device_address, int handle, const char* data, unsigned short data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Parameter eintragen
        // Tabellen Handle
        *reinterpret_cast<int *>(&m_packet_data[count]) = handle;
        count += sizeof(int);
        // Daten
        std::memcpy( &m_packet_data[count], data, data_size );
        count += data_size;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA_1, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA_1, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA_1, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::write_current_table_row_data_ext( unsigned char device_address, int handle, unsigned short flags, char separator, const char* data, unsigned short data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, %c, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, flags, separator, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Wenn weder die flags noch der Seperator verwendet wird, auf die Vorgaengerversion der Funktion zugreifen.
    if (flags == 0 && separator == 0)
        return write_current_table_row_data(device_address, handle, data, data_size);

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Parameter eintragen
        // Tabellen Handle
        *reinterpret_cast<int *>(&m_packet_data[count]) = handle;
        count += sizeof(int);
        // Flags
        *reinterpret_cast<unsigned short*>(&m_packet_data[count]) = flags;
        count += sizeof(unsigned short);
        // Separator
        m_packet_data[count] = static_cast<unsigned char>(separator);
        count++;
        // Daten
        std::memcpy( &m_packet_data[count], data, data_size );
        count += data_size;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_ROW_DATA, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::write_current_table_column_data( unsigned char device_address, int handle, const char* name, const char* data )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, device_address, handle, name, data );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Name als Parameter eintragen
        *reinterpret_cast<int *>( m_packet_data ) = handle;
        count = sizeof(int);
        count += sprintf(reinterpret_cast<char *>(&m_packet_data[count]), "%s", name) + 1;
        count += sprintf(reinterpret_cast<char *>(&m_packet_data[count]), "%s", data) + 1;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA_1, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA_1, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA_1, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::write_current_table_column_data_ext( unsigned char device_address, int handle, unsigned short flags, const char* name, const char* data, unsigned short data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, %i, 0x%08X, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, flags, name, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Wenn die flags nicht verwendet werden, auf die Vorgaengerversion der Funktion zugreifen.
    if (flags == 0)
        return write_current_table_column_data(device_address, handle, name, data);

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );


        // Name als Parameter eintragen
        *reinterpret_cast<int *>( m_packet_data ) = handle;
        count = sizeof(int);
        *reinterpret_cast<unsigned short *>( &m_packet_data[count] ) = flags;
        count += sizeof(unsigned short);
        count += sprintf( reinterpret_cast<char *>( &m_packet_data[count] ), "%s", name ) + 1;
        memcpy(&m_packet_data[count], data, data_size);
        count += data_size;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_CURRENT_COLUMN_DATA, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::write_all_table_rows_column_data( unsigned char device_address, int handle, unsigned short flags, const char* name, const char* data, unsigned short data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, %i, 0x%08X, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, flags, name, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );


        // Name als Parameter eintragen
        *reinterpret_cast<int *>( m_packet_data ) = handle;
        count = sizeof(int);
        *reinterpret_cast<unsigned short *>( &m_packet_data[count] ) = flags;
        count += sizeof(unsigned short);
        count += sprintf( reinterpret_cast<char *>( &m_packet_data[count] ), "%s", name ) + 1;
        memcpy(&m_packet_data[count], data, data_size);
        count += data_size;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktive Antwort abwarten
        if ( (result = call( IV_CONTEXT_SLAVE, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA_OK, m_packet_data, count, IV_PENDING_TRIALS( get_connection().get_read_timeout() ) )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA_OK, result );
            return result;
        }

    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::reply_write_all_table_rows_column_data( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        trace( "%s(%i):ERROR(1): %s", __FILE__, __LINE__, __func__ );
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_SET_ALL_ROWS_TO_COLUMN_DATA_OK, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_current_table_row_data( unsigned char device_address, int handle, char* data, unsigned short &data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Handle als Parameter eintragen
        *reinterpret_cast<int *>( &m_packet_data[count] ) = handle;
        count += sizeof(int);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA_1, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA_1, result );
            return result;
        }

        // OK
        if ( count > 256 )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        if ( count > data_size )
        {
            trace( "%s(%i):WARNING(1): %s [ Puffer für Antwortdaten zu klein gewählt. ]\n", __FILE__, __LINE__, __func__, count );
            // Wert uebernehmen
            std::copy_n( m_packet_data, data_size, data );
        }
        else
        {
            // Wert uebernehmen
            std::copy_n( reinterpret_cast<char *>( m_packet_data ), count, data );
            
            data_size = count;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA_1, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_current_table_row_data_ext( unsigned char device_address, int handle, unsigned short flags, char separator, char* data, unsigned short &data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, %c, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, flags, separator, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Wenn weder die flags noch der Seperator verwendet wird, auf die Vorgaengerversion der Funktion zugreifen.
    if (flags == 0 && separator == 0)
        return read_current_table_row_data(device_address, handle, data, data_size);

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Handle als Parameter eintragen
        *reinterpret_cast<int *>( &m_packet_data[count] ) = handle;
        count += sizeof(int);
        // Flags
        *reinterpret_cast<unsigned short*>(&m_packet_data[count]) = flags;
        count += sizeof(unsigned short);
        // Separator
        m_packet_data[count] = static_cast<unsigned char>(separator);
        count++;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA, result );
            return result;
        }

        // OK
        if ( count > 256 )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        if ( count > data_size )
        {
            trace( "%s(%i):WARNING(1): %s [ Puffer für Antwortdaten zu klein gewählt. ]\n", __FILE__, __LINE__, __func__, count );
            // Wert uebernehmen
            memcpy( data, m_packet_data, data_size );
        }
        else
        {
            // Wert uebernehmen
            memcpy( data, reinterpret_cast<char *>( m_packet_data ), count );
            data_size = count;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_ROW_DATA, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_current_table_column_data( unsigned char device_address, int handle, const char* name, char* data, unsigned short &data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, name, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Handle als Parameter eintragen
        *reinterpret_cast<int *>( m_packet_data ) = handle;
        count = sizeof( int );
        count += sprintf( reinterpret_cast<char *>( &m_packet_data[count] ), "%s", name ) + 1;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA_1, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA_1, result );
            return result;
        }

        // OK
        if ( count > 440 ) // Binaerfeld mit 220 Byte als Hexstring.
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
        if (count > data_size)
        {
            trace( "%s(%i):WARNING(1): %s [ Puffer für Antwortdaten zu klein gewählt. ]\n", __FILE__, __LINE__, __func__, count );
            // Wert uebernehmen
            memcpy(data, m_packet_data, data_size);
        }
        else
        {
            // Wert uebernehmen
            memcpy(data, m_packet_data, count);
            data_size = count;
            if (data[data_size] == 0)
                data_size--; // Nullterminie abziehen.

        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_current_table_column_data_ext( unsigned char device_address, int handle, unsigned short flags, const char* name, char* data, unsigned short &data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, name, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Wenn die flags nicht verwendet werden, auf die Vorgaengerversion der Funktion zugreifen.
    if (flags == 0)
        return read_current_table_column_data(device_address, handle, name, data, data_size);

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Handle als Parameter eintragen
        *reinterpret_cast<int *>( m_packet_data ) = handle;
        count = sizeof(int);
        *reinterpret_cast<unsigned short *>( &m_packet_data[count] ) = flags;
        count += sizeof(unsigned short);
        count += sprintf( reinterpret_cast<char *>( &m_packet_data[count] ), "%s", name ) + 1;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA, result );
            return result;
        }

        // OK
        if ( count > 440 ) // Binaerfeld mit 220 Byte als Hexstring.
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
        if (count > data_size)
        {
            trace( "%s(%i):WARNING(1): %s [ Puffer für Antwortdaten zu klein gewählt. ]\n", __FILE__, __LINE__, __func__, count );
            // Wert uebernehmen
            memcpy(data, m_packet_data, data_size);
        }
        else
        {
            // Wert uebernehmen
            memcpy(data, m_packet_data, count);
            data_size = count;

        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GET_CURRENT_COLUMN_DATA, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::append_table_row_data( unsigned char device_address, int handle, unsigned short flags, char separator, const char* data, unsigned short data_size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, %c, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, handle, flags, separator, data, data_size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Parameter eintragen
        // Tabellen Handle
        *reinterpret_cast<int *>(&m_packet_data[count]) = handle;
        count += sizeof(int);
        // Flags
        *reinterpret_cast<unsigned short*>(&m_packet_data[count]) = flags;
        count += sizeof(unsigned short);
        // Separator
        m_packet_data[count] = static_cast<unsigned char>(separator);
        count++;
        // Daten
        std::memcpy( &m_packet_data[count], data, data_size );
        count += data_size;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_APPEND_ROW_DATA, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_APPEND_ROW_DATA, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_APPEND_ROW_DATA, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::delete_current_table_row(unsigned char device_address, int handle, unsigned short flags)
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, , 0x%08X ]\n", __FILE__, __LINE__, __func__, device_address, handle, flags);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Handle als Parameter eintragen
        *reinterpret_cast<int *>( &m_packet_data[count] ) = handle;
        count = sizeof( int );
        // Flags
        *reinterpret_cast<unsigned short*>(&m_packet_data[count]) = flags;
        count += sizeof(unsigned short);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_CURRENT_ROW, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_CURRENT_ROW, result );
            return result;
        }

        // OK
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

    }
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_CURRENT_ROW, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::delete_filtrated_table_rows(unsigned char device_address, int handle, unsigned short flags)
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, , 0x%08X ]\n", __FILE__, __LINE__, __func__, device_address, handle, flags);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Handle als Parameter eintragen
        *reinterpret_cast<int *>( &m_packet_data[count] ) = handle;
        count = sizeof( int );
        // Flags
        *reinterpret_cast<unsigned short*>(&m_packet_data[count]) = flags;
        count += sizeof(unsigned short);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS, result );
            return result;
        }

        // OK
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktive Antwort abwarten
        if ( (result = call( IV_CONTEXT_SLAVE, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS_OK, m_packet_data, count, IV_PENDING_TRIALS( get_connection().get_read_timeout() ) )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS_OK, result );
            return result;
        }
    }
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::reply_delete_filtrated_table_rows( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace( "%s(%i):ERROR(1): %s", __FILE__, __LINE__, __func__ );
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_DELETE_AVAILABLE_ROWS_OK, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::append_fingerprint_record( unsigned char device_address, unsigned char type, const unsigned char* data, int size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, type, data, size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    int i;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );


        // Parameter eintragen
        m_packet_data[count] = type;
        count = sizeof( unsigned char );

        for ( i = 0; i < size; ++i )
        {
            m_packet_data[count + i] = data[i];
        }
        count += size;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD, result );
            return result;
        }

        // OK
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktive Antwort abwarten
        if ( (result = call( IV_CONTEXT_SLAVE, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD_OK, m_packet_data, count, IV_PENDING_FINGERPRINT_TRIALS( get_connection().get_read_timeout() ) )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD_OK, result );
            return result;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::reply_append_fingerprint_record( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace( "%s(%i):ERROR(1): %s", __FILE__, __LINE__, __func__ );
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_APPEND_RECORD_OK, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::read_fingerprint_record( unsigned char device_address, unsigned char &type, unsigned int pid, unsigned char fid, unsigned char* data, int &size )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, %i, %i, 0x%08X, %i ]\n", __FILE__, __LINE__, __func__, device_address, type, pid, fid, data, size );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0, i;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Parameter eintragen
        m_packet_data[count] = type;
        count = sizeof( unsigned char );
        *reinterpret_cast<unsigned int *>( &m_packet_data[count] ) = pid;
        count += sizeof( unsigned int );
        m_packet_data[count] = fid;
        count += sizeof( unsigned char );

        if ( (result = call(IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Wenn die Funktion unbekannt ist,
            // wird die Vorgaengerversion ohne "Aktive Antwort" geprueft.
            if ( ((result >> 8) & 0xFF) == 2 )
            {
                if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD1, m_packet_data, count )) != IV_ERROR_SUCCESS )
                {
                    // Fehler ausgeben
                    trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD1, result );
                    return result;
                }

                // Templatetype uebernehmen
                type = m_packet_data[0];
                // Templatedaten uebernehmen
                size = count - 1;
                for ( i = 1; i < count; ++i )
                {
                    data[i - 1] = m_packet_data[i];
                }
                count += size;

                // OK zurueckmelden
                result = IV_ERROR_SUCCESS;
                return result;
            }


            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD, result );
            return result;
        }

        // OK
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktive Antwort abwarten
        if ( (result = call( IV_CONTEXT_SLAVE, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD_OK, m_packet_data, count, IV_PENDING_FINGERPRINT_TRIALS( get_connection().get_read_timeout() ) )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD_OK, result );
            return result;
        }

        // Templatetype uebernehmen
        type = m_packet_data[0];
        // Templatedaten uebernehmen
        size = count - 1;
        // Templatedaten muessen mindestens 15 Byte haben.
        if ( size < 15 )
            return -IV_ERROR_PROCESSING;

        for ( i = 1; i < count; ++i )
        {
            data[i - 1] = m_packet_data[i];
        }
        count += size;
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::reply_read_fingerprint_record( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace( "%s(%i):ERROR(1): %s", __FILE__, __LINE__, __func__ );
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_GET_RECORD_OK, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::delete_fingerprint_record( unsigned char device_address, unsigned int pid, unsigned char fid )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, pid, fid );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Parameter eintragen
        *reinterpret_cast<unsigned int *>( &m_packet_data[count] ) = pid;
        count = sizeof( unsigned int );
        m_packet_data[count] = fid;
        count += sizeof( unsigned char );

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Wenn die Funktion unbekannt ist,
            // wird die Vorgaengerversion ohne "Aktive Antwort" geprueft.
            if ( ((result >> 8) & 0xFF) == 2 )
            {
                // Vorgaengerfunktion ausfuehren.
                if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD1, m_packet_data, count )) != IV_ERROR_SUCCESS )
                {
                    // Fehler ausgeben
                    trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD1, result );
                    return result;
                }

                // OK
                if ( count )
                {
                    trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
                }

                // OK zurueckmelden
                result = IV_ERROR_SUCCESS;
                return result;
            }

            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD, result );
            return result;
        }

        // OK
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten.  ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktive Antwort abwarten
        if ( (result = call( IV_CONTEXT_SLAVE, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD_OK, m_packet_data, count, IV_PENDING_FINGERPRINT_TRIALS( get_connection().get_read_timeout() ) )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD_OK, result );
            return result;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::reply_delete_fingerprint_record( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace( "%s(%i):ERROR(1): %s", __FILE__, __LINE__, __func__ );
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_FINGERPRINT_TABLE_DELETE_RECORD_OK, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::generate_table( unsigned char device_address, const std::string &name, const std::string &reserved )
{
    trace( "%s(%i):INFO(3): %s [ %i, %s, %s ]\n", __FILE__, __LINE__, __func__, device_address, name.c_str(), reserved.c_str() );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver( device_address );

        // Alle Zusatzparameter uebernehmen
        std::copy(name.cbegin(), name.cend(), &m_packet_data[count]);
        count += static_cast<int>( name.size() );
        m_packet_data[count] = 0; // Nullterminierung
        count++;

        std::copy(reserved.cbegin(), reserved.cend(), &m_packet_data[count]);
        count += static_cast<int>( reserved.size() );
        m_packet_data[count] = 0; // Nullterminierung
        count++;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GENERATE, m_packet_data, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUNCTION_TABLE_GENERATE, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count )
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count );
        }
        else
        {
            count = 0;
        }

        // Aktive Antwort abwarten
        if ( (result = call( IV_CONTEXT_SLAVE, IV_MODULE_TABLE, IV_FUCNTION_TABLE_GENERATE_OK, m_packet_data, count, IV_PENDING_FINGERPRINT_GENERATE_TRIALS( get_connection().get_read_timeout() ) )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            trace_error( get_device_address_receiver(), IV_MODULE_TABLE, IV_FUCNTION_TABLE_GENERATE_OK, result );
            return result;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUNCTION_TABLE_GENERATE, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_table::reply_generate_table( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace( "%s(%i):ERROR(1): %s", __FILE__, __LINE__, __func__ );
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_TABLE, IV_FUCNTION_TABLE_GENERATE_OK, tmp, count )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return result;
}
