/**
* @file  file_device.cpp
* @brief Implementierung der Klasse FileDevice.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_device.h"
#include "common/software_info.h"
#include <time.h>
#include <fstream>
#include <cstring>

using namespace Datafox;

FileDevice::FileDevice() : FileIntelHex()
    , _fileHeader( FileIdDeviceUpdate )
{
    // Initalisieren
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::device_header );
}

FileDevice::~FileDevice()
{
    // Speicher freigeben.
    resetData();
}

void FileDevice::resetData()
{
    // Basis Reset
    FileIntelHex::resetData();

    // Header loeschen
    m_buffer.resize( sizeof( SDFFileDeviceHeader) );
    std::fill( m_buffer.begin(), m_buffer.end(), '\0' );
    // Hilfsstrukturzeiger
    m_header = reinterpret_cast<SDFFileDeviceHeader *>( &m_buffer[0] );

    // Firmwaredaten loeschen
    m_fileRawFirmware.resetData();

    // Anzeigedaten loeschen
    m_fileDisplayFonts.resetData();
    m_fileDisplayGrafics.resetData();

    // Geraetetexte loeschen
    m_fileDeviceText.resetData();
}

unsigned int FileDevice::GetHeaderVersionFile()
{ 
    return m_header->versionFile;
}

unsigned int FileDevice::GetHeaderVersionTimestamp( std::string &timestamp )
{
    timestamp = boost::str( boost::format( "%1%" ) % std::string( m_header->timestamp ) );
    return static_cast<unsigned int>( timestamp.size() );
}

unsigned int FileDevice::GetHeaderVersionFirmware( std::string &versionFirmware )
{
    versionFirmware = boost::str( boost::format( "%1%" ) % std::string( m_header->versionFirmware ) );
    return static_cast<unsigned int>( versionFirmware.size() );
}

SoftwarebuildVersion FileDevice::GetHeaderVersionFirmware()
{
    return SoftwarebuildVersion( m_header->versionFirmware );
}

StructSoftwareInfo FileDevice::GetHeaderSoftwareInfo()
{
    return m_header->softwareInfo;
}

SDFFileDeviceHeader FileDevice::GetHeader( const std::string &filePath )
{
    // Fehlerobjekt zurücksetzen.
    resetError();

    // Lesen der Hexdatei
    FILE *fp;
    char data[0x400];

    unsigned char headerBuf[IV_DFFILEDEVICE_SIZE_HEADER];
    memset(&headerBuf, 0, sizeof(headerBuf));
    SDFFileDeviceHeader *header = (SDFFileDeviceHeader*)headerBuf;

    if ( (fp = fopen(filePath.data(), "r")) == (FILE*)0 )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return *header;
    }

    unsigned int numLine = 0;
    unsigned char count;
    unsigned short i, address;

    while( !feof( fp ) && (errorState() == Error_state::none) )
    {
        if ( fgets( data, 0x400, fp ) != 0 )
        {
            if ( static_cast<Record_type>( GetType( data ) ) == Record_type::device_header )
            {
                address = GetAddress( data );
                count   = GetCount( data );
                // spaetere Groessenanderung vorbehalten!
                for ( i = address; (i < IV_DFFILEDEVICE_SIZE_HEADER) && ((i - address) < count); ++i )
                    headerBuf[i] = GetData8( data, i - address );
            }

            numLine++;
        }
    }

    fclose( fp );	

    // Abgeaenderter Header ab 03.01.33 ist groesser 4 Byte
    if ( header->sizeData ) // Vorher war nur die Version vorhanden!
    {
        // CRC16 von Header pruefen!
        // Der crcHead ist immer der letzte 16bit Wert in der Struktur und
        // geht ueber den gesamten Dateiheader, sich selbst ausgeschlossen.
        if ( *(unsigned short*)&headerBuf[header->sizeData - sizeof(unsigned short)] != GetCRC16( headerBuf, header->sizeData - sizeof(unsigned short) ) )
        {
            // Fehler zuweisen
            setErrorState( Error_state::head_crc );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

            memset( &headerBuf, 0, sizeof(headerBuf) );
            return *header;
        }

        // Aeltere Struktur auf aktuell verwendete Version anpassen.
        if ( header->sizeData < sizeof(SDFFileDeviceHeader) )
        {
            // CRC-Position hat eine neue Bedeutung, vorsichtshalber hier löschen.
            *(unsigned short*)&headerBuf[header->sizeData - sizeof(unsigned short)] = 0;

            //############################################
            // SoftwareInfo ggf. auf die aktuelle Version anpassen.
            header->softwareInfo = CheckSoftwareInfo( header->softwareInfo );

            // CRC nach hinten verschieben, jedoch die Struktur-Version (sizeData) beibehalten.
            header->crcHead = GetCRC16( headerBuf, header->sizeData - sizeof(unsigned short) );
        }
        else if ( header->sizeData > sizeof(SDFFileDeviceHeader) )
        {
            // Fehler zuweisen
            setErrorState( Error_state::version );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

            memset( &headerBuf, 0, sizeof(headerBuf) );
            return *header;
        }
    }
    else
    {
        // Fehler zuweisen
        setErrorState( Error_state::function_unavailable );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

        memset( &headerBuf, 0, sizeof(headerBuf) );
        return *header;
    }

    return *header;
}

int FileDevice::CheckVersionFile( SDFFileDeviceHeader &header )
{
    // Version auswerten (Majornumber muss uebereinstimmen) Erst ab: 03.01.33
    if ( ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_PZE   & 0xFFFF0000))  
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_ZK    & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_BDE   & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MOBIL & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MDE   & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_AE        & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TIMEBOY   & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MULTIPLE  & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_FLEX      & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_EXKLUSIV  & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TIMEBOY_MOBIL_PZE & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TERMINAL_2_8 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TERMINAL_4_3 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_5_7 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_10 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_Q7 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IO_BOX & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_ZK_BOX & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MOBIL_BOX & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_SOFTWARE_INFO & 0xFFFF0000)) )
    {
        // Fehler zuweisen
        setErrorState( Error_state::device_unknown );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    return 1;
}

unsigned int FileDevice::GetVersionFile( const std::string &filePath )
{
    SDFFileDeviceHeader header = GetHeader( filePath );

    if ( !header.sizeData || !CheckVersionFile( header ) )
        return 0;

    return header.versionFile;
}

unsigned int FileDevice::GetVersionFirmware( const std::string &filePath, std::string &versionFirmware )
{
    SDFFileDeviceHeader header = GetHeader( filePath );

    if ( !header.sizeData || !CheckVersionFile( header ) )
        return 0;

    versionFirmware = boost::str( boost::format( "%1%" ) % std::string( header.versionFirmware ) );
    return static_cast<unsigned int>( versionFirmware.size() );
}

bool FileDevice::load( const std::string &filePath ) 
{
    bool res = true;
    // Datenumgebung zuruecksetzen.
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

    // Header lesen
    *m_header = GetHeader( filePath.data() );

    if ( !m_header->sizeData || !CheckVersionFile( *m_header ) )
    {
        // Header loeschen
        memset( m_headBuf, 0, sizeof(m_headBuf) );
        m_header = (SDFFileDeviceHeader*)m_headBuf;
        return false;
    }

    // Dateiname halten
    setFilePath( filePath );

    // Firmwaredaten laden
    if ( !m_fileRawFirmware.load( filePath ) )
        return false;

    unsigned int first = m_fileRawFirmware.firstAddress();
    unsigned int last = m_fileRawFirmware.lastAddress();
    std::size_t size = m_fileRawFirmware.sizeData();

    // CRC32 von Program-Flash pruefen
    if ( m_header->crcProgram != m_fileRawFirmware.checksum32bit() )
    {
        // Fehler zuweisen
        setErrorState( Error_state::program_crc );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return false;
    }

    // Displaydaten Zeichensaetze laden
    if ( !m_fileDisplayFonts.load( filePath ) )
    {
        // Keine Displaydaten vorhanden?
        if ( Error_state::no_data != m_fileDisplayFonts.errorState() )
        {
            setErrorState( Error_state::load_display_data );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }
    }
    // Displaydaten Grafiken laden
    if ( !m_fileDisplayGrafics.load( filePath ) )
    {
        // Keine Displaydaten vorhanden?
        if ( Error_state::no_data != m_fileDisplayGrafics.errorState() )
        {
            setErrorState( Error_state::load_display_data );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }
    }
    // Geraetetexte laden
    if ( !m_fileDeviceText.load( filePath ) )
    {
        // Keine Textdaten vorhanden?
        if ( Error_state::no_data != m_fileDeviceText.errorState() )
        {
            setErrorState( Error_state::load_device_text );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }
    }

    return res;
}

bool FileDevice::loadBin( const std::string &filePath )
{
    std::ifstream fs( filePath, std::ios::binary );

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( fs.is_open() == false )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        // Dateigroesse ermitteln.
        fs.seekg( 0, fs.end );
        std::ifstream::pos_type size = fs.tellg();
        fs.seekg( 0, fs.beg );

        // Dateiheader lesen und auswerten.
        FileHeader::DataType structHeader;

        if ( size > fileHeader().size() )
        {
            fs.read( reinterpret_cast<char *>( structHeader.data() ), structHeader.size() );

            StructFileHeader &structFileHeader = reinterpret_cast<StructFileHeader &>( structHeader );
            // Auf Dateiheader prüfen
            if( fileHeader().data().df[0] != structFileHeader.df[0] || fileHeader().data().df[1] != structFileHeader.df[1] || fileHeader().data().id != structFileHeader.id )
            {
                // Keine gueltige Dateiversionsangabe vorhanden.
                setErrorState( Error_state::version_file );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
            else
            {
                size -= fileHeader().size();
                // Puffer anpassen.
                bufferBin_.resize( static_cast<std::vector<unsigned char>::size_type>( size ) );
                // Daten laden
                fs.read( reinterpret_cast<char *>( bufferBin_.data() ), size );
            }
        }

        if ( errorState() == Error_state::none )
        {
            // Member durch gelesene Daten initialisieren und pruefen
            if ( bufferBin_.size() >= sizeof(StructDeviceUpdateNode) )
            {
                const StructDeviceUpdateNode &mvn = reinterpret_cast<const StructDeviceUpdateNode &>( *bufferBin_.data() );
                if ( mvn.header.id != EnumStructId::StructIdDeviceUpdateNode )
                {
                    // Keine gueltige Dateiversionsangabe vorhanden.
                    setErrorState( Error_state::version_file );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                }
                else if ( mvn.header.id == EnumStructId::StructIdDeviceUpdateNode && mvn.header.size > sizeof(StructDeviceUpdateNode) )
                {
                    // Version der Dateidaten aktueller als erwartet.
                    setErrorState( Error_state::version );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                }
                else
                {
                    // Aus den Serialisierungsdaten die Baumrepresentation erstellen.
                    readSerialize();
                }
            }
            else
            {
                // Fehler zuweisen
                setErrorState( Error_state::no_data );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
        }
    }

    return (errorState() == Error_state::none);
}

/*!
Hilfsfunktion zum manuellen Serialisieren der in _memHardwareConfig enthaltenen Daten.
*/
bool FileDevice::readSerialize()
{
    // Aus den Serialisierungsdaten die Baumrepresentation erstellen.
    if ( !_deviceUpdateNode.readSerialize( bufferBin_ ) )
    {
        // Erstellen der Dateidaten gescheitert
        setErrorState( Error_state::serialize_file_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {

        // Header loeschen
        memset( m_headBuf, 0, sizeof(m_headBuf) );
        m_header = (SDFFileDeviceHeader*)m_headBuf;
        // Header zusammenstellen.
        const StructDeviceUpdateSegmentFirmware &dusf = _deviceUpdateNode.getSegmentFirmware();
        m_header->versionFile = IV_DFFILEDEVICE_VERSION_SOFTWARE_INFO;
        m_header->sizeData = sizeof(SDFFileDeviceHeader);
        std::strncpy( m_header->timestamp, dusf.timestamp, sizeof(m_header->timestamp) );

        m_fileRawFirmware.setContent( _deviceUpdateNode.getSegmentDataFirmware() );
    }

    return (errorState() == Error_state::none);
}

bool FileDevice::save( const std::string &filePath, SDFFileDeviceHeader &header ) 
{
    // Fehlerobjekt zuruecksetzen
    resetError();

    //############################
    // Headerangaben pruefen

    // Eingetragene Strukturgroesse sicherstellen
    if ( !header.sizeData )
        header.sizeData = sizeof(SDFFileDeviceHeader);

    // Dateiversionsangabe
    if ( !CheckVersionFile( header ) )
        return false;

    // Firmwareversion pruefen
    if ( (strlen( header.versionFirmware ) < 8) || (strlen( header.versionFirmware ) > 32) )
    {
        // Fehler zuweisen
        setErrorState( Error_state::firmware_version_length );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return false;
    }

    // Builddatum, wenn nicht vorhanden, eintragen
    if ( strlen( header.timestamp ) != 19 )
    {
        time_t ltime;
        struct tm *ltm;

        time( &ltime );
        ltm = localtime( &ltime );
        strftime( header.timestamp, sizeof(header.timestamp), "%Y-%m-%d %H:%M:%S", ltm );
    }

    // CRC32 fuer Program-Flash berechnen
    header.crcProgram = m_fileRawFirmware.checksum32bit();

    // CRC16 fuer Dateiheader berechnen
    unsigned char *pCRC = &((unsigned char*)&header)[header.sizeData - sizeof(unsigned short)];
    *(unsigned short*)pCRC = GetCRC16( (unsigned char*)&header, header.sizeData - sizeof(unsigned short) );


    //############################
    // Schreiben der Hexdatei

    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::device_header );
    m_buffer.resize( sizeof( SDFFileDeviceHeader) );
    *reinterpret_cast<SDFFileDeviceHeader *>( &m_buffer[0] ) = header;

    if ( FileIntelHex::save( filePath, false, false ) )
    {
        // Firmwaredaten schreiben
        if ( !m_fileRawFirmware.save( filePath, true, false ) )
            return false;

        // Anzeigedaten in Datei schreiben
        if ( m_fileDisplayFonts.sizeData() )
        {
            if ( !m_fileDisplayFonts.save( filePath, true, false ) )
                return false;
        }

        // Anzeigedaten in Datei schreiben
        if ( m_fileDisplayGrafics.sizeData() )
        {
            if ( !m_fileDisplayGrafics.save( filePath, true, false ) )
                return false;
        }

        // Geraetetexte
        if ( m_fileDeviceText.sizeData() )
        {
            if ( !m_fileDeviceText.save( filePath, true, false ) )
                return false;
        }

        // Schreiben der internen Speicher in die Hex-Datei.
        std::ofstream fileOut( filePath, std::ios::app );
        std::string lineBuffer( endLine() );
        if ( lineBuffer.size() == 0 )
        {
            // Fehler zuweisen
            setErrorState( Error_state::record_end_make );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
        else
        {
            fileOut << lineBuffer << std::endl;
        }
    }

    return (errorState() == Error_state::none);
}

bool FileDevice::saveBin( const std::string &filePath, SDFFileDeviceHeader &header )
{
    std::ofstream fs( filePath, std::ios::binary );

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( fs.is_open() == false )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        //############################
        // Headerangaben pruefen

        // Eingetragene Strukturgroesse sicherstellen
        if ( !header.sizeData )
            header.sizeData = sizeof(SDFFileDeviceHeader);

        // Dateiversionsangabe
        if ( !CheckVersionFile( header ) )
            return false;

        // Firmwareversion pruefen
        if ( (strlen( header.versionFirmware ) < 8) || (strlen( header.versionFirmware ) > 32) )
        {
            // Fehler zuweisen
            setErrorState( Error_state::firmware_version_length );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }

        // Builddatum, wenn nicht vorhanden, eintragen
        if ( strlen( header.timestamp ) != 19 )
        {
            time_t ltime;
            struct tm *ltm;

            time( &ltime );
            ltm = localtime( &ltime );
            strftime( header.timestamp, sizeof(header.timestamp), "%Y-%m-%d %H:%M:%S", ltm );
        }

        // CRC32 fuer Program-Flash berechnen
        header.crcProgram = m_fileRawFirmware.checksum32bit();

        // CRC16 fuer Dateiheader berechnen
        unsigned char *pCRC = &((unsigned char*)&header)[header.sizeData - sizeof(unsigned short)];
        *(unsigned short*)pCRC = GetCRC16( (unsigned char*)&header, header.sizeData - sizeof(unsigned short) );

        // Serialisierung der Baumstruktur in das zu schreibende Array durchfuehren.
        if ( !writeSerialize( header ) )
            return false;

        // Dateiheader schreiben
        fs.write( reinterpret_cast<const char *>( &fileHeader().data() ), fileHeader().size() );
        // Daten schreiben
        fs.write( reinterpret_cast<const char *>( bufferBin_.data() ), bufferBin_.size() );
    }

    return (errorState() == Error_state::none);
}

/*!
Hilfsfunktion zum manuellen Serialisieren der in _memHardwareConfig enthaltenen Daten.
*/
bool FileDevice::writeSerialize( const SDFFileDeviceHeader &header )
{
    // Aktuellen Speicher leeren.
    bufferBin_.clear();
    // DeviceUpdateNode neu aufsetzen.
    _deviceUpdateNode.resetData();
    // Vorhandenen Segmente hinzufügen.
    StructDeviceUpdateSegmentFirmware dusf;
    std::fill_n(reinterpret_cast<unsigned char*>(&dusf), sizeof(dusf), 0);
    dusf.header.id = StructIdDeviceUpdateSegmentFirmware;
    dusf.header.size = sizeof(StructDeviceUpdateSegmentFirmware);
    dusf.size = (unsigned int)m_fileRawFirmware.sizeData();
    dusf.crc32 = header.crcProgram;
    std::strncpy( dusf.timestamp, header.timestamp, sizeof(dusf.timestamp) - 1 );
    std::strncpy( dusf.version, header.versionFirmware, sizeof(dusf.version) - 1 );
    dusf.softwareInfo = header.softwareInfo;
    // Segmentheader inkl. Daten übergeben.
    _deviceUpdateNode.setSegmentFirmware( dusf, m_fileRawFirmware.content() );
    // Ggf. Segment hinzufügen.
    if ( m_fileDisplayFonts.sizeData() )
    {
        StructDeviceUpdateSegment dus;
        dus.header.id = StructIdDeviceUpdateSegment;
        dus.header.size = sizeof(StructDeviceUpdateSegment);
        dus.type = 3;
        dus.size = (unsigned int)m_fileDisplayFonts.sizeData();
        dus.crc32 = m_fileDisplayFonts.checksum32bit();
        _deviceUpdateNode.appendSegment( dus, m_fileDisplayFonts.content() );
    }
    // Ggf. Segment hinzufügen.
    if ( m_fileDisplayGrafics.sizeData() )
    {
        StructDeviceUpdateSegment dus;
        dus.header.id = StructIdDeviceUpdateSegment;
        dus.header.size = sizeof(StructDeviceUpdateSegment);
        dus.type = 4;
        dus.size = (unsigned int)m_fileDisplayGrafics.sizeData();
        dus.crc32 = m_fileDisplayGrafics.checksum32bit();
        _deviceUpdateNode.appendSegment( dus, m_fileDisplayGrafics.content() );
    }
    // Ggf. Segment hinzufügen.
    if ( m_fileDeviceText.sizeData() )
    {
        StructDeviceUpdateSegment dus;
        dus.header.id = StructIdDeviceUpdateSegment;
        dus.header.size = sizeof(StructDeviceUpdateSegment);
        dus.type = 11;
        dus.size = (unsigned int)m_fileDeviceText.sizeData();
        dus.crc32 = m_fileDeviceText.checksum32bit();
        _deviceUpdateNode.appendSegment( dus, m_fileDeviceText.content() );
    }
    // Serialisierung der Baumstruktur in das zu schreibende Array durchfuehren.
    if ( !_deviceUpdateNode.writeSerialize( bufferBin_ ) )
    {
        // Erstellen der Dateidaten gescheitert
        setErrorState( Error_state::serialize_file_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }

    return (errorState() == Error_state::none);
}

int FileDevice::build( const std::string &filePath, SDFFileDeviceHeader &header, const std::string &fileFirmware, const std::string &fileDisplayFonts, const std::string &fileDisplayGrafics, const std::string &fileDeviceText )
{
    // Umgebungsreset
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

    // Laden der Firmware
    if ( !m_fileRawFirmware.load( fileFirmware ) )
        return 0;

    if ( fileDisplayFonts.size() )
    {
        // Laden der Anzeigedaten
        if ( !m_fileDisplayFonts.load( fileDisplayFonts ) )
            return 0;
    }

    if ( fileDisplayGrafics.size() )
    {
        // Laden der Anzeigedaten
        if ( !m_fileDisplayGrafics.load( fileDisplayGrafics ) )
            return 0;
    }

    if ( fileDeviceText.size() )
    {
        // Importieren der Geraetetexte
        if ( !m_fileDeviceText.Import( fileDeviceText ) )
            return 0;
    }

    // Schreiben der Geraetedatei
    if ( !save( filePath, header ) )
        return 0;

    return 1;
}

StructSoftwareInfo FileDevice::SoftwareInfoStringToStruct( const std::string &str )
{
    std::string tmp( str );
    StructSoftwareInfo swInfo;
    memset( &swInfo, 0, sizeof(StructSoftwareInfo) );

    // # durch : ersetzen
    tmp[INTELHEX_OFFSET_START] = INTELHEX_RECORD_MARK;    // start of record
    // CRC pruefen
    if ( !checkCrc( tmp.data() ) )
        return swInfo;

    unsigned char i, count = GetCount( tmp.data() );
    // spaetere Groessenanderung vorbehalten!
    for ( i = 0; (i < count) && i < sizeof(StructSoftwareInfo); ++i )
        ((unsigned char*)&swInfo)[i] = GetData8(tmp.data(), i);

    return CheckSoftwareInfo( swInfo );
}

bool FileDevice::SoftwareInfoStructToString( std::string &str, StructSoftwareInfo &swInfo )
{
    bool res = false;

    str = dataLine( 0, reinterpret_cast<unsigned char *>( &swInfo ), sizeof(StructSoftwareInfo) );
    if ( str.length() )
    {
        res = true;
        // : durch # ersetzen
        str[INTELHEX_OFFSET_START] = '#';    // start of record
    }

    return res;
}

StructSoftwareInfo FileDevice::CheckSoftwareInfo( const StructSoftwareInfo &swInfo )
{
    return _deviceInfo.checkInfo( swInfo );
}

FileIntelHex *FileDevice::GetFileRawFirmware()
{
    return &m_fileRawFirmware;
}

FileDisplayFonts *FileDevice::GetFileDisplayFonts()
{
    return &m_fileDisplayFonts;
}

FileDisplayGrafics *FileDevice::GetFileDisplayGrafics()
{
    return &m_fileDisplayGrafics;
}

FileDeviceText *FileDevice::GetFileDeviceText()
{
    return &m_fileDeviceText;
}

int FileDevice::CompareInfos( const StructHardwareInfo &hwInfo, const StructSoftwareInfo &swInfo )
{
    return _deviceInfo.compareInfos( hwInfo, swInfo );
}

int FileDevice::SoftwareInfoToString( const StructSoftwareInfo &swInfo, std::string &info )
{
    if ( !_deviceInfo.SoftwareInfoToString( swInfo, info ) )
    {
        setErrorText( _deviceInfo.m_errorText );
        return false;
    }

    return true;
}

